/***************************************************************************
    DbxConv - DBX to MBOX Converter

    author:  Ulrich Krebs
    email :  ukrebs@freenet.de

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream.h>
#include <direct.h>
#include <afx.h>
#include "DbxReader.h"

#define OPT_NO 0
#define OPT_MBOXO 1
#define OPT_MBOXRD 2
#define OPT_DOUBLENEWLINE 4
#define OPT_IGNORECASE 8
#define OPT_FLATFOLDERS 16
#define OPT_EML 32
#define OPT_USERECEIVEDATE 64
#define OPT_USESENDDATE 128
#define OPT_OVERWRITEEXISTING 256
#define OPT_HELP 4096

CString FormatFilename(CDbxReader *dbx, int msgNr, CString format)
{
	CString outString="";
	CString varName;
	CString undoBuf;
	int startPos=0,endPos=0;

	while((startPos=format.Find("$",0))>=0)
	{
		if((endPos=format.Find("$",startPos+1))<=startPos) break;
		outString+=format.Left(startPos);
		varName=format.Mid(startPos,endPos-startPos+1);
		undoBuf=varName;
		format=format.Right(format.GetLength()-endPos-1);

		varName=varName.Left(varName.GetLength()-1);
		varName=varName.Right(varName.GetLength()-1);
		CString tokens[5];
		int index=0;
		int tokenPos;

		while(index<4 && (tokenPos=varName.Find("_",0))>=0)
		{
			tokens[index++]=varName.Left(tokenPos);
			varName=varName.Right(varName.GetLength()-tokenPos-1);
		}
		tokens[index]=varName;

		CString s;
		int parsePos=0;
		bool parseFound=false;
		bool parseOK=false;

		int maxLength=-1;
		int capitalization=-1;
		int numItems=-1;
		CString empty;
		CString dateFormat="%Y-%m-%d";
		
		while(parsePos<=index)
		{
			if(tokens[parsePos].Find("L:")==0)
			{
				maxLength=atoi(tokens[parsePos].Mid(2));
			}
			else if(tokens[parsePos].Find("C:")==0)
			{
				capitalization=atoi(tokens[parsePos].Mid(2));
			}
			else if(tokens[parsePos].Find("E:")==0)
			{
				empty=tokens[parsePos].Mid(2);
			}
			else if(tokens[parsePos].Find("F:")==0)
			{
				dateFormat=tokens[parsePos].Mid(2);
			}
			else if(tokens[parsePos].Find("N:")==0)
			{
				numItems=atoi(tokens[parsePos].Mid(2));
			}
			parsePos++;
		}

		if(tokens[0]=="DBXNAME")
		{
			int pos;
			s=dbx->GetFileName();
			if((pos=s.ReverseFind('.'))>=0) s=s.Left(pos);
			parseOK=true;
		}
		else if(tokens[0]=="DBXDATE")
		{
			s=dbx->GetFileDate().Format(dateFormat);
			parseOK=true;
		}
		else
		{
			if(msgNr>=0)
			{
				if(tokens[0]=="SADDR")
				{
					s=dbx->GetSenderAddress(msgNr);
					parseOK=true;
				}
				else if(tokens[0]=="RADDR")
				{
					s=dbx->GetReceiverAddress(msgNr);
					parseOK=true;
				}
				else if(tokens[0]=="SNAME")
				{
					s=dbx->GetSender(msgNr);
					parseOK=true;
				}
				else if(tokens[0]=="RNAME")
				{
					s=dbx->GetReceiver(msgNr);
					parseOK=true;
				}
				else if(tokens[0]=="SUBJ")
				{
					s=dbx->GetSubject(msgNr);
					parseOK=true;
				}
				else if(tokens[0]=="RDATE")
				{
					s=dbx->GetReceiveDate(msgNr).Format(dateFormat);
					parseOK=true;
				}
				else if(tokens[0]=="SDATE")
				{
					s=dbx->GetSendDate(msgNr).Format(dateFormat);
					parseOK=true;
				}
			}
		}

		if(parseOK)
		{
			if(numItems>0 && (tokens[0]=="RADDR" || tokens[0]=="RNAME"))
			{
				CString buf=s;
				int pos=0;
				int item=0;
				while(pos<buf.GetLength())
				{
					if(buf[pos]==';')
					{
						item++;
						if(item==numItems) break;
					}
					pos++;
				}
				s=buf.Mid(0,pos);
			}

			if(empty!="" && s=="") s=empty;
			if(maxLength>0 && maxLength<256) s=s.Left(maxLength);
			if(capitalization==0) s.MakeLower();
			else if(capitalization==1) s.MakeUpper();
			outString+=s;
		}
		else
		{
			outString+=undoBuf.Left(undoBuf.GetLength()-1);
			format="$"+format;
			continue;
		}
	}
	outString+=format;

	CString notAllowed("<>|?/\\\":*\r\n");
	int i;
	
	for(i=0;i<outString.GetLength();i++)
	{ 
		if((unsigned char)outString.GetAt(i) < ' ')  outString.SetAt(i,' ');
	}


	for(i=0;i<notAllowed.GetLength();i++)
	{
		outString.Replace(notAllowed.GetAt(i),' ');
	}
	
	while(outString.Replace("  "," ")>0);

	outString.TrimLeft();
	outString.TrimRight();

	return outString;
}


CString ReplaceFrom(CString s, int method)
{
	int findPos=0;

	do
	{
		if(findPos>=s.GetLength()) break;

		if(method&OPT_MBOXRD)
		{
			while(s[findPos]=='>') findPos++;
		}
		if(method&OPT_IGNORECASE)
		{
			if(s.Mid(findPos,5).CompareNoCase("From ")==0) s.Insert(findPos,">");
		}
		else
		{
			if(s.Mid(findPos,5).Compare("From ")==0) s.Insert(findPos,">");
		}
	}
	while((findPos=s.Find('\n',findPos)+1)!=0);
	return s;
}


int main(int argc, char* argv[])
{
	int options=OPT_NO;
	CString fileSpec;
	CString outDir;
	CString mbxFilenameFormat="$DBXNAME$.mbx";
	CString emlFilenameFormat="$SNAME_L:32_E:Unknown$ - $SUBJ_L:64_E:No Subject$.eml";
	int i;

	enum
	{
		EXPECT_NONE=0,
		EXPECT_OPTION=1,
		EXPECT_INFILE=2,
		EXPECT_OUTDIR=4
	};

	int expect=EXPECT_OPTION|EXPECT_INFILE;

	for(i=1;i<argc;i++)
	{
		CString s(argv[i]);

		if(expect&EXPECT_OPTION && s[0]=='-')
		{
			if(s.CompareNoCase("-dn")==0)
			{
				options|=OPT_DOUBLENEWLINE;
				continue;
			}
			if(s.CompareNoCase("-ic")==0)
			{
				options|=OPT_IGNORECASE;
				continue;
			}
			if(s.CompareNoCase("-ff")==0)
			{
				options|=OPT_FLATFOLDERS;
				continue;
			}
			if(s.Left(6).CompareNoCase("-mbxrd")==0)
			{
				options|=OPT_MBOXRD;
				if(s.GetLength()>6) mbxFilenameFormat=s.Mid(6);
				continue;
			}
			if(s.Left(4).CompareNoCase("-mbx")==0)
			{
				options|=OPT_MBOXO;
				if(s.GetLength()>4) mbxFilenameFormat=s.Mid(4);
				continue;
			}
			if(s.Left(4).CompareNoCase("-eml")==0)
			{
				options|=OPT_EML;
				if(s.GetLength()>4) emlFilenameFormat=s.Mid(4);
				continue;
			}
			if(s.CompareNoCase("-rcvdate")==0)
			{
				options|=OPT_USERECEIVEDATE;
				continue;
			}
			if(s.CompareNoCase("-senddate")==0)
			{
				options|=OPT_USESENDDATE;
				continue;
			}
			if(s.CompareNoCase("-overwrite")==0)
			{
				options|=OPT_OVERWRITEEXISTING;
				continue;
			}
			if(s.CompareNoCase("-?")==0)
			{
				options|=OPT_HELP;
				break;
			}
			printf("Unknown option %s!\n",s);
			return 1;
		}

		if(expect&EXPECT_INFILE)
		{
			fileSpec=s;
			expect=EXPECT_OUTDIR;
			continue;
		}

		if(expect&EXPECT_OUTDIR)
		{
			outDir=s;
			expect=EXPECT_NONE;
			continue;
		}
	}

	if(argc==1 || options&OPT_HELP)
	{
		printf("DBX to MBOX and EML Converter, Ver. 1.3.4, by Ulrich Krebs\n\n");
		printf("Usage: dbxconv [Options] Infile [Output directory]\n\n");
		printf("Available options:\n");
		printf("  -mbx[naming]   Saves all messages in mboxo-format (default)\n");
		printf("  -mbxrd[naming] Saves all messages in mboxrd-format\n");
		printf("  -dn            Always add a double newline at the end of a message\n");
		printf("  -ic            Ignore case when replacing 'From'\n");
		printf("  -ff            Flat folders, do not create subdirectories for eml files\n");
		printf("  -eml[naming]   Saves all messages in eml-format\n");
		printf("                 By default, all messages will be saved into a\n");
		printf("                 subdirectory named as the dbx-file.\n");
		printf("                 Use -ff to store all messages into one folder.\n");
		printf("  -senddate      Uses send date as timestamp (default)\n");
		printf("  -rcvdate       Uses receive date as timestamp\n");
		printf("  -overwrite     Overwrite existing messages with same name\n");
		printf("  -?             Shows this message\n");
		return 0;
	}

	if(fileSpec.GetLength()==0)
	{
		printf("You must specify a filename!\n");
		return 1;
	}

	CStringArray filePaths;
	CFileFind finder;
	BOOL found=finder.FindFile(fileSpec);
	while(found)
	{
		found=finder.FindNextFile();
		if(!finder.IsDirectory())
		{
			filePaths.Add(finder.GetFilePath());
		}
	}

	if(filePaths.GetSize()==0)
	{
		printf("Nothing to do!\n");
		return 0;
	}

	if(!(options&OPT_MBOXO) && !(options&OPT_MBOXRD) && !(options&OPT_EML))
	{
		printf("No conversion options selected, using MBOXO-converter...\n");
		options|=OPT_MBOXO;
	}

	int mboxSuccess=0;
	int emlSuccess=0;
	CString filePath;
	CString fileName;
	CString oemFileName;
	CString outFileDir;
	CString outFileName;
	CString outFilePath;
	CString oemOutFilePath;

	for(i=0;i<filePaths.GetSize();i++)
	{
		CDbxReader *dbx;
		CFile *inFile;
		CFile *outFile;

		filePath=filePaths.GetAt(i);

		TRY
		{
			inFile=new CFile(filePath,CFile::modeRead);
		}
		CATCH(CFileException, e)
		{
			printf("Unable to open file \"%s\"!\n",filePath);
			continue;
		}
		END_CATCH

		fileName=inFile->GetFileName();
		oemFileName=fileName;
		oemFileName.AnsiToOem();

		TRY
		{
			dbx=new CDbxReader(inFile);
		}
		CATCH(CFileException, e)
		{
			printf("File \"%s\" seems to be defective!\n",oemFileName);
			delete(inFile);
			continue;
		}
		END_CATCH

		if(dbx->GetType()!=dbx->TYPE_EMAIL)
		{
			if(dbx->GetType()==dbx->TYPE_OE4)
			{
				printf("Unable to handle Outlook Express mailboxes below ver. 5!\n");
			}
			else printf("\"%s\" is not an Outlook Express mailbox!\n",oemFileName);
			delete(dbx);
			delete(inFile);
			continue;
		}

		if(dbx->GetItemCount()==0)
		{
			printf("\"%s\" contains no messages, skipped...\n",oemFileName);
			delete(dbx);
			delete(inFile);
			continue;
		}

		printf("Processing mailbox \"%s\" (%ld messages)...\n",oemFileName,dbx->GetItemCount());
		
		if(options&OPT_MBOXO || options&OPT_MBOXRD)
		{
			int pos;
			outFileName=FormatFilename(dbx, -1, mbxFilenameFormat);

			outFilePath="";
			if(outDir.GetLength()==0)
			{
				outFilePath=filePath;
				if((pos=outFilePath.ReverseFind('\\'))!=-1)
				{
					outFilePath=outFilePath.Left(pos);
				}
				if(outFilePath.Right(1)!="\\") outFilePath+="\\";
				outFilePath=outFilePath+outFileName;
			}
			else
			{
				if(outDir.Right(1)!="\\") outDir+="\\";
				outFilePath=outDir+outFileName;
			}

			oemOutFilePath=outFilePath;
			oemOutFilePath.AnsiToOem();

			TRY
			{
				outFile=new CFile(outFilePath,CFile::modeWrite|CFile::modeCreate|CFile::typeBinary);
			}
			CATCH(CFileException, e)
			{
				printf("Unable to open outputfile \"%s\"!\n",oemOutFilePath);
				outFile=NULL;
				delete(dbx);
				delete(inFile);
				continue;
			}
			END_CATCH

			TRY
			{
				for(int j=0;j<dbx->GetItemCount();j++)
				{
					CString s;
					CTime t;
					if(options&OPT_USERECEIVEDATE)
					{
						t=dbx->GetReceiveDate(j);
					}
					else
					{
						t=dbx->GetSendDate(j);
					}
					CString ts=t.Format("%a %b %d %X %Y");
					CString sender=dbx->GetSenderAddress(j);
					if(sender.IsEmpty()) sender="-";
					s="From "+sender+" "+ts;
					outFile->Write(s,s.GetLength());
					outFile->Write("\r\n",2);
					s=dbx->GetMessage(j);
					s=ReplaceFrom(s,options);
					outFile->Write(s,s.GetLength());
					outFile->Write("\r\n",2);
					if(options&OPT_DOUBLENEWLINE || (s.GetLength()>0 && s[s.GetLength()-1]!='\n'))
					{
						outFile->Write("\r\n",2);
					}
				}
				mboxSuccess++;
				outFile->Close();
				delete(outFile);
			}
			CATCH(CFileException, e)
			{
				if(e->m_cause==CFileException::badSeek || e->m_cause==CFileException::endOfFile || e->m_cause==CFileException::generic)
				{
					printf("File \"%s\" seems to be defective!\n",oemFileName);
				}
				else
				{
					printf("Unable to write to outputfile \"%s\" (disk full?)\n",oemOutFilePath);
				}
				outFile->Close();
				delete(outFile);
//			The output file is sometimes at least partly ok, so I'll leave it up to the
//			user to take a look at it
//			CFile::Remove(outFilePath);
			}
			END_CATCH
		}
		
		if(options&OPT_EML)
		{
			int pos;
			outFileDir="";
			if(outDir.GetLength()==0)
			{
				outFileDir=filePath;
				if((pos=outFileDir.ReverseFind('\\'))!=-1)
				{
					outFileDir=outFileDir.Left(pos);
				}
			}
			else
			{
				outFileDir=outDir;
			}
			if(outFileDir.Right(1)!="\\") outFileDir+="\\";

			if(!(options&OPT_FLATFOLDERS))
			{
				if((pos=fileName.ReverseFind('.'))!=-1)
				{
					outFileDir+=fileName.Left(pos);
				}
				else
				{
					outFileDir+=fileName;
				}
				if(outFileDir.Right(1)!="\\") outFileDir+="\\";
			}

			mkdir(outFileDir);

			int processedMails=0;
			for(int j=0;j<dbx->GetItemCount();j++)
			{
				outFileName=FormatFilename(dbx, j, emlFilenameFormat);
				outFilePath=outFileDir+outFileName;

				if(!(options&OPT_OVERWRITEEXISTING))
				{
					int k=1;
					CFile testFile;
					char buf[16];
					while(testFile.Open(outFilePath,CFile::modeRead))
					{
						testFile.Close();
						if((pos=outFileName.ReverseFind('.'))!=-1)
						{
							outFilePath=outFileDir+outFileName.Left(pos)+"("+itoa(k,buf,10)+")"+outFileName.Mid(pos);
						}
						else
						{
							outFilePath=outFileDir+outFileName+"("+itoa(k,buf,10)+")";
						}
						k++;
					}
				}

				oemOutFilePath=outFilePath;
				oemOutFilePath.AnsiToOem();
					
				TRY
				{
					outFile=new CFile(outFilePath,CFile::modeWrite|CFile::modeCreate|CFile::typeBinary);
				}
				CATCH(CFileException, e)
				{
					printf("Unable to open outputfile \"%s\"!\n",oemOutFilePath);
					outFile=NULL;
					break;
				}
				END_CATCH

				TRY
				{
					CString s=dbx->GetMessage(j);
					outFile->Write(s,s.GetLength());
				}
				CATCH(CFileException, e)
				{
					if(e->m_cause==CFileException::badSeek || e->m_cause==CFileException::endOfFile || e->m_cause==CFileException::generic)
					{
						printf("File \"%s\" seems to be defective!\n",oemFileName);
					}
					else
					{
						printf("Unable to write to outputfile \"%s\" (disk full?)\n",oemOutFilePath);
					}
					outFile->Close();
					delete(outFile);
					CFile::Remove(outFilePath);
					break;
				}
				END_CATCH
				outFile->Close();
				delete(outFile);
				CFileStatus fileStatus;
				CFile::GetStatus(outFilePath,fileStatus);
				if(options&OPT_USERECEIVEDATE)
				{
					fileStatus.m_mtime=dbx->GetReceiveDate(j);
				}
				else
				{
					fileStatus.m_mtime=dbx->GetSendDate(j);
				}
				CFile::SetStatus(outFilePath,fileStatus);
				processedMails++;
			}
			if(processedMails==dbx->GetItemCount())
			{
				emlSuccess++;
			}
		}

		inFile->Close();

		delete(dbx);
		delete(inFile);
	}
	
	if(options&OPT_MBOXO)
	{
		printf("%d of %d mailboxes converted to MBOXO-format!\n",mboxSuccess,filePaths.GetSize());
	}
	if(options&OPT_MBOXRD)
	{
		printf("%d of %d mailboxes converted to MBOXRD-format!\n",mboxSuccess,filePaths.GetSize());
	}
	if(options&OPT_EML)
	{
		printf("%d of %d mailboxes converted to EML-format!\n",emlSuccess,filePaths.GetSize());
	}

	return 0;
}
